/****
KrigApprox Implementation - 8/2/00 cmsief
****/
#include "KrigApprox.h"

/**************************CONSTRUCTORS + DESTRUCTOR***************************/
  
KrigApprox::KrigApprox():Approximate(){
  v=NULL;
  theta=NULL; beta=NULL;
  sigma2=0; PE=NULL;
  estimation_trend=TREND_CONSTANT;
  custom_f=NULL;MLE=NULL;fdim=1;first_time=true;
}/*end KrigApprox*/

KrigApprox::KrigApprox(long dim):Approximate(dim){
/*Call SetDefault*/
  v=NULL; theta=NULL;beta=NULL;PE=NULL;MLE=NULL;first_time=true;
  SetDefault(dim);  
}/*end special constructor*/
  
KrigApprox::KrigApprox(KrigApprox &K):Approximate(K){
  sigma2=K.sigma2;estimation_trend=K.estimation_trend;
  if(K.v!=NULL) v=new mp_vector(*K.v);
  else v=NULL;
  if(K.theta!=NULL) theta=new mp_vector(*K.theta);
  else theta=NULL;
  if(K.beta!=NULL) beta=new mp_vector(*K.beta);
  else beta=NULL;
  if(K.PE!=NULL) PE=new ParamEstimate(*K.PE);
  else PE=NULL;
  if(K.MLE!=NULL) MLE=new pt_collect(*K.MLE);
  else MLE=NULL;
  custom_f=(K.custom_f);
  fdim=K.fdim;first_time=K.first_time;
}/*end copy constructor*/  

KrigApprox::KrigApprox(ParamEstimate &PE1, long dim) {
  SetDefault(dim);first_time=true;
  (*PE)=PE1;
  (*theta).newsize((*PE).GetThetaSize());
}/*end special constructor*/

KrigApprox::~KrigApprox() {
  if(PE!=NULL) delete PE;
  if(v!=NULL) delete v;
  if(theta!=NULL) delete theta;
  if(beta!=NULL) delete beta;
  if(MLE!=NULL) delete MLE;
}/*end destructor*/

KrigApprox& KrigApprox::operator= (KrigApprox &K) {
  if(this!=&K) {

    if(PE!=NULL) delete PE;
    if(v!=NULL) delete v;
    if(theta!=NULL) delete theta;
    if(beta!=NULL) delete beta;

    P=K.P;sigma2=K.sigma2;estimation_trend=K.estimation_trend;
    custom_f=K.custom_f;fdim=K.fdim;first_time=K.first_time;
    
    if(K.PE!=NULL) PE=new ParamEstimate(*K.PE);
    else PE=NULL;
    if(K.v!=NULL) v=new mp_vector(*K.v);
    else v=NULL;
    if(K.theta!=NULL) theta=new mp_vector(*K.theta);
    else theta=NULL;
    if(K.beta!=NULL) beta=new mp_vector(*K.beta);
    else beta=NULL;
    if(K.MLE!=NULL) MLE=new pt_collect(*K.MLE);
    else MLE=NULL;    

  }/*end if*/
  return (*this);
}/*end overloaded =*/  

/***************************RESET FUNCTIONS***********************************/

void KrigApprox::ResetKrigApproxState(ParamEstimate &PE1) {
/****
INPUT: ParamEstimate object.
EFFECT: Resets this object's state.
****/
  (*PE)=PE1;  
}/*end ResetKrigApproxState*/
  
bool KrigApprox::ResetEstimateConstantTrend() {
/****
OUTPUT: Always returns true.
EFFECT: Chooses Estimation with a constant trend.  Since the trend is a
constant, no parameter is needed.
****/
  estimation_trend=TREND_CONSTANT;fdim=1;
  return true;
}/*end ResetEstimateConstantTrend*/
  
bool KrigApprox::ResetEstimateQuadraticTrend() {
/****
OUTPUT: Always returns true.
EFFECT: Chooses Estimation with that quadratic trend.
****/
  estimation_trend=TREND_QUADRATIC;
  fdim=1 + P + P*(P+1)/2;
  return true;  
}/*end ResetEstimateQuadraticTrend*/
  
  
bool KrigApprox::ResetEstimateCustomTrend(void(*fcn)(const mp_vector &,
                                          mp_vector&), long fcn_dim){
/****
INPUT: A function pointer, the the vector-valued trend function.
OUTPUT: Always returns true.
EFFECT: Chooses Custom trend Estimation.
****/
  custom_f=fcn;
  estimation_trend=TREND_CUSTOM;
  fdim=fcn_dim;
  return true;
}/*end ResetEstimateConstantTrend*/
  
bool KrigApprox::ResetCorrelationFamily(long cor) {
/****
INPUT: Correlation family number (see ParamEstimate.h).
OUTPUT: true if correlation family number is valid.
EFFECT: return PE.ResetCorrelationFamily(cor);
****/
  bool s;
  if(PE!=NULL) {
    s=(*PE).ResetCorrelationFamily(cor);
    if (s) (*theta).newsize((*PE).GetThetaSize());
    return s;
  }/*end if*/
  else return false;
}/*end ResetCorrelationFamily*/
  
/********************************QUERY FUNCTIONS*******************************/

void KrigApprox::debug(char *code) const{
/****
EFFECT: Outputs a state dump to stdout.
****/
  cout<<code<<":***KrigApprox::debug()***\n";
  cout<<code<<":P="<<P<<" sigma2="<<sigma2<<" fdim="<<fdim<<endl;
  if(estimation_trend==TREND_CONSTANT)
      cout<<code<<":Constant Trend Estimation\n";
  else if(estimation_trend==TREND_QUADRATIC)
      cout<<code<<":Quadratic Trend Estimation\n";
  else if(estimation_trend==TREND_CUSTOM)
      cout<<code<<":Custom Trend Estimation\n";
  else
      cout<<code<<":Trend unknown\n";
  if(PE!=NULL)
      (*PE).debug(code);
  else
      cout<<code<<":PE=NULL\n";
  if(v!=NULL)
      cout<<code<<":v="<<(*v);
  else
      cout<<code<<":v=NULL\n";
  if(beta!=NULL)
      cout<<code<<":beta="<<(*beta);
  else
      cout<<code<<":beta=NULL\n"; 
  if(theta!=NULL)
      cout<<code<<":theta="<<(*theta);
  else
      cout<<code<<":theta=NULL\n";
  if(MLE!=NULL)
      cout<<code<<":MLE="<<(*MLE);
  else
      cout<<code<<":MLE=NULL\n";
  cout<<code<<":***end KA::debug()***\n";
}/*end debug*/
  
/*******************************ACTION FUNCTIONS*******************************/
  
bool KrigApprox::EvaluateApprox(const mp_vector &x, const pt_collect &Eval_c,
                                double* Approx, double* D_c){
/****
INPUT: The location to evaluate at, the location of the previously
evaluated sites, and two double pointers for return values.
OUTPUT: Returns true if the operation was successful.
EFFECT: Using kriging to evaluate the approximation at this point.  It also
evaluates the design criterion.
****/
#if DEBUG>=3
  cout<<"KAEA   :function called\n";
  cout<<"KAEA   :X="<<x;
#endif

  /* Return false if the approximation has not yet been built using
     UpdateApprox */
  if(first_time) return false;

  long points=(*v).dim();
  long i; /*counters*/
  mp_vector r(points);
  mp_vector MLE_r(points+fdim);

  /*Create the correlation vector between the points and the design sites*/
  for(i=0;i<fdim;i++) MLE_r[i]=1;
  for(i=0;i<points;i++) {
    r[i]=(*PE).EvaluateMyCorrelation(P,x,Eval_c.row(i), (*theta));
    MLE_r[i+fdim]=r[i];
  }/*end for*/
#if DEBUG>=4
  cout<<"KAEA   : r="<<r;
  cout<<"KAEA   : MLE_r="<<MLE_r;
#endif

  /*evaluate the value of the approximation here*/
  (*Approx)=EvaluateTrend(x) + (*v) * r;

  /*evaluate the sqrt of the mean squared error for the design criterion*/
  double MLE_X = MLE_r * ((*MLE) * MLE_r); /*uses hacked multiply*/
#if DEBUG>=3
  cout<<"KAEA   :MLE_X="<<MLE_X<<endl;
#endif

  /* Make design criterion zero if errors were encountered */
  /* sigma2 was calculated in ParamEstimate and passed back in
     UpdateApprox */
  if(sigma2<0) {
#if DEBUG>=2
      cout<<"KAEA : ERROR(non-fatal) sigma2="
          << sigma2 << " Zeroing D_C..." << endl; (*D_c)=0;
#endif
  }/*end if*/
  else if((1-MLE_X)<0)
    {
#if DEBUG>=2
      cout<<"KAEA   : ERROR(non-fatal) 1-MLE_X="
          << 1-MLE_X << " Zeroing D_C..." << endl;
      cout<<"KAEA   : ERROR(non-fatal) x="<<x;
#endif
      (*D_c)=0;
    }/*end if*/
  else
  (*D_c) = sqrt(sigma2 * ( 1.0 - MLE_X));/*uses hacked multiply to see
                                         first MLE_r as a row vector*/
  
#if DEBUG>=3
  cout<<"KAEA   :function completed\n";
#endif
  return true;
  
}/*end EvaluateApprox*/
  
bool KrigApprox::UpdateApprox(pt_collect &Eval_c, mp_vector&
                              f_values, long numpoints,
                              double delta, const Bounds &b){
/****
INPUT: The collection of evaluated sites, their values, the number of points
evaluated, the iteration number, delta and the bounds.
OUTPUT: Returns true if the operation was successful.
EFFECT: This updates the weighting scheme and then rekrigs the function,
calling for parameter estimation as needed.  Also, it updates the Design
Criterion.
****/
  if(MLE==NULL) MLE=new pt_collect(1,1);
  (*MLE).newsize(fdim+numpoints, fdim+numpoints);
  if(v==NULL) v=new mp_vector(1,1);  
  (*v).newsize(numpoints);

  bool s=false;

  if(first_time) {
    first_time=false;
    if(theta==NULL) theta=new mp_vector((*PE).GetThetaSize());    

    /*Get an initial value for theta*/
    double L=b.upper[0]-b.lower[0];
    for(long i=1;i<P;i++)
      if(b.upper[i]-b.lower[i] > L) L=b.upper[i]-b.lower[i];
      
    L=-100*log(0.5) / sqr(L);

    /*Verify bounds on initial theta*/
    if(L>=PE->GetThetaUpperBound(delta))
      L=PE->GetThetaUpperBound(delta)/2.0;
    (*theta)=L;
  }/*end if*/
  
  switch(estimation_trend) {
  case TREND_CONSTANT:
      s=(*PE).EstimateConstantTrend(Eval_c, f_values, delta,
                                    (*beta).begin(), &sigma2,
                                    theta, MLE, v);
      break;
  case TREND_QUADRATIC:
      s=(*PE).EstimateQuadraticTrend(Eval_c, f_values, delta, beta,&sigma2,
                                     theta, MLE, v);
      break;
  case TREND_CUSTOM:
      s=(*PE).EstimateCustomTrend(Eval_c, f_values, delta, custom_f, fdim,
                                  beta, &sigma2, theta, MLE, v);
      break;
  default:
      s=false;    
  };/*end switch*/
  return s;  
}/*end UpdateApprox*/

void KrigApprox::SetDefault(long dim) {
/****
INPUT: Dimension of space.  Used for parameter estimation.
EFFECT: Constant Trend Estimation, and ParamEstimate's default family.
****/
  PE=new ParamEstimate(dim);
  P=dim;
  v=new mp_vector(1);/*it needs to be numpoints later*/
  theta=new mp_vector((*PE).GetThetaSize());
  
  (*theta)=0.0001; /*assigns a guess to all elements of theta*/
  estimation_trend=TREND_CONSTANT;
  beta=new mp_vector(1);/*Assumes Trend Constant*/
  sigma2=0;
  fdim=1;
  custom_f=NULL;   
  MLE=new pt_collect(1,1);/*will be resized later*/
}/*end SetDefault*/


double KrigApprox::EvaluateTrend(const mp_vector& x){
/****
INPUT: A mp_vector, x 
OUTPUT: The value of the trend at that particular point.  This is assembled
from the betas, as appropriate.
****/
  mp_vector tempv(fdim);  
  switch(estimation_trend) {
  case TREND_CONSTANT:
    return (*beta)[0];/*for constant trend, value is in beta[0]*/
    break;
  case TREND_QUADRATIC:
    /*Code from A. Padula's gls.cc*/
    long j, g, h;     
    tempv[0] = 1;
    g = P+1;
    for( j = 1; j <=P; j++) {
      tempv[j] = x[j-1];
      
      for(h = j-1;h < P; h++) {
        tempv[g] = x[j-1]*x[h];
        g++;
      }/*end for*/
    }/*end for*/
    return tempv * (*beta);
    break;
  case TREND_CUSTOM:
    (*custom_f)(x,tempv);
    return tempv * (*beta);/*hacked multpliy assumes transpose*/
    break;
  default:
    return ERROR;    
  };/*end switch*/

}/*end EvaluateTrend*/







