/****
KrigApprox Class - 8/2/00 cmsief
Derived from Approximate.  This class is a kriging-based function approximation.
NOTE: You must call Update before Evaluate the first time.
****/
#ifndef _MAPS_KRIG_APPROX_
#define _MAPS_KRIG_APPROX_

#include "maps_general.h"
#include "ParamEstimate.h"
#include "Approximate.h"

/*estimation trend codes*/
#define TREND_CONSTANT 0
#define TREND_QUADRATIC 1
#define TREND_CUSTOM 2

class KrigApprox: public Approximate {

public:
  KrigApprox();
  KrigApprox(long dim);
  /*Call SetDefault*/
  KrigApprox(KrigApprox &K);
  KrigApprox(ParamEstimate &PE1, long dim);
  virtual ~KrigApprox();

  KrigApprox& operator= (Approximate &A)
    {if(A.GetType()==A_KRIG_APPROX) *this=(KrigApprox&)A;
     return *this;}
  
  KrigApprox& operator= (KrigApprox &K);  
  
  /*****RESET FUNCTIONS*****/

  void ResetKrigApproxState(ParamEstimate &PE1);
  /****
  INPUT: ParamEstimate object.
  EFFECT: Resets this object's state.
  ****/
  
  bool ResetEstimateConstantTrend();
  /****
  OUTPUT: Always returns true.
  EFFECT: Chooses Estimation with a constant trend.  Since the trend is a
  constant, no parameter is needed.
  ****/
  
  bool ResetEstimateQuadraticTrend();
  /****
  OUTPUT: Always returns true.
  EFFECT: Chooses Estimation with that quadratic trend.
  ****/
  
  bool ResetEstimateCustomTrend(void(*fcn)(const mp_vector &, mp_vector&), long
                                fcn_dim);
  /****
  INPUT: A function pointer, the vector-valued trend function and its dimension.
  OUTPUT: Always returns true.
  EFFECT: Chooses Custom trend Estimation.
  ****/

  bool ResetCorrelationFamily(long cor);
  /****
  INPUT: Correlation family number (see ParamEstimate.h).
  OUTPUT: true if correlation family number is valid.
  EFFECT: return PE.ResetCorrelationFamily(cor);
  ****/
  
  /*****QUERY FUNCTIONS*****/

  long GetType() const {return A_KRIG_APPROX;}
  /****
  OUTPUT: returns the proper code for type of approximation
  ****/     
  
  void debug(char *code) const;
  /****
  EFFECT: Outputs a state dump to stdout.
  ****/

  /*****ACTION FUNCTIONS*****/
  
  bool EvaluateApprox(const mp_vector &x, const pt_collect &Eval_c,
                      double* Approx, double* D_c);
  /****
  INPUT: The location to evaluate at, the location of the previously
  evaluated sites, and two double pointers for return values.
  OUTPUT: Returns true if the operation was successful.
  EFFECT: Using kriging to evaluate the approximation at this point.  It also
  evaluates the design criterion.
  ****/
  
  bool UpdateApprox(pt_collect &Eval_c, mp_vector &f_values, long
                    numpoints, double delta,const Bounds &b);
  /****
  INPUT: The collection of evaluated sites, their values, the number of points
  evaluated, the iteration number, the current delta and the bounds;
  OUTPUT: Returns true if the operation was successful.
  EFFECT: This updates the weighting scheme and then rekrigs the function,
  calling for parameter estimation as needed.  Also, it updates the Design
  Criterion.
  ****/

private:

  void SetDefault(long dim);
  /****
  INPUT: Dimension of space.  Used for parameter estimation.
  EFFECT: Constant Trend Estimation, and ParamEstimate's default family.
  ****/ 

  double EvaluateTrend(const mp_vector& x); 
  /****
  INPUT: A mp_vector, x 
  OUTPUT: The value of the trend at that particular point.  This is assembled
  from the betas, as appropriate.
  ****/
  
  mp_vector* v;        /*from estimation*/
  mp_vector* theta;    /*Estimated*/
  mp_vector* beta;     /*Estimated*/
  double sigma2;       /*Estimated*/
  ParamEstimate* PE;
  void(*custom_f)(const mp_vector &, mp_vector&);/*For custom function estimate*/
  int estimation_trend;/*see top for codes*/
  pt_collect* MLE;     /*the MLE matrix... from estimation*/
  long fdim;           /*dimension of the vector-valued function fcn*/
  bool first_time;
};/*end class*/

#endif
