/****
PSGrid implementation - 6/19/00 cmsief
MAPS uses this class to find the core pattern around a point and to find
the best start step for the optimization.
****/

#include "PSGrid.h"
#include <fstream>
#include <iostream>
/***********************CONSTRUCTORS + DESTRUCTOR**************************/
PSGrid::PSGrid(){
  P=0;
  origin=NULL;
  gridvect=NULL;
  refine=0.0;
}/*end default construtor*/

PSGrid::PSGrid(const PSGrid &g){
  origin=NULL;
  gridvect=NULL;
  ResetPSGridState(g.P,*g.origin,*g.gridvect, g.refine, g.delta);
}/*end copy constructor*/

PSGrid::PSGrid(long P1, const mp_vector& o, const pt_collect& g,
               double ref, double d){
  origin=NULL;
  gridvect=NULL;
  ResetPSGridState(P1,o,g,ref,d);
}/*end special constructor*/

PSGrid::PSGrid(long P1, const Bounds &b){
/*this assumes a certain relative grid thickness*/
  double minres;

  mp_vector resolution(P1);
  resolution=PSG_DEFAULT_RESOLUTION * (b.upper-b.lower);

  /*Makes sure resolution is constant across dimensions*/
  minres=resolution[0];
  for(long i=1;i<P1;i++)
    if(resolution[i]<minres) minres=resolution[i];

  resolution=minres;

  origin=NULL;
  gridvect=NULL;
  SetDefault(P1,b,resolution);
}/*end special constructor*/

PSGrid::PSGrid(long P1, const Bounds &b, const mp_vector &resolution) {
  origin=NULL;
  gridvect=NULL;
  SetDefault(P1, b, resolution);
}/*end special constructor*/

PSGrid::~PSGrid(){
  if(origin !=NULL) delete origin;
  if(gridvect!=NULL) delete gridvect;
}/*end destructor*/

PSGrid& PSGrid::operator=(PSGrid &g) {
  if(this!=&g) {
    P=g.P;refine=g.refine;delta=g.delta;
    if(origin !=NULL) delete origin;
    if(g.origin==NULL) origin=NULL;
    else origin=new mp_vector(*g.origin);

    if(gridvect !=NULL) delete gridvect;
    if(g.gridvect==NULL) gridvect=NULL;
    else gridvect=new pt_collect(*g.gridvect);    
  }
  return (*this);
}/*end overloaded =*/

/*****************************RESET FUNCTIONS********************************/


void PSGrid::ResetPSGridState(long P1, const mp_vector& o, const pt_collect& g,
                              double ref, double d){
/****
INPUT: A complete set of state values for the grid
EFFECTS: It copies these into the object's state
****/
  P=P1;
  refine=ref;
  delta=d;
  /*Memory Allocation+Deallocation*/
  if(origin!=NULL) delete origin;
  if(gridvect!=NULL)  delete gridvect;
  
  origin=new mp_vector(P);
  gridvect=new pt_collect(P,P);/*this makes sure there is enough memory*/

  (*origin)=o;
  (*gridvect)=g;
}/*end ResetState*/
  
bool PSGrid::ReadGridFromFile(char* fn){
/****
INPUT: File name string
EFFECTS: Reads the grid data from a file
FILE FORMAT: dimension<cr> 
             delta<cr>
             refine<cr>
             origin(P)<cr>
             grid(PxP)
****/
  ifstream ifs;
  long dim,i,j;
  double ref,d;
  ifs.open(fn,ios::in);

  ifs>>dim;
  if(ifs.eof()) return false;
  ifs>>d;
  if(ifs.eof()) return false; 
  ifs>>ref;
  if(ifs.eof()) return false; 
  
  /*Memory allocation*/
  mp_vector* zor=new mp_vector(dim);
  pt_collect* grid=new pt_collect(dim,dim);

  /*origin*/
  for(i=0;i<dim&&!ifs.eof();i++) ifs>>(*zor)[i];
  if(ifs.eof()) {ifs.close(); delete zor; delete grid; return false;}

  /*gridvect*/
  for(i=0;i<dim&&!ifs.eof();i++)
    for(j=0;j<dim&&!ifs.eof();j++)
      ifs>>((*grid)[i])[j];

  /*final checks*/
  if(ifs.eof()) {
    ifs.close();
    delete zor; delete grid;
    return false;
  }/*end if*/
  else {
    ifs.close();
    P=dim;
    refine=ref;
    delta=d;
    delete origin;
    delete gridvect;
    origin=zor;
    gridvect=grid;
    return true;
  }/*end else*/
}/*end ReadGridFromFile*/

/*******************************QUERY FUNCTIONS*******************************/

void PSGrid::debug(char *code) const{
/****
EFFECTS: does a complete state dump to stdout.
****/
  cout<<code<<":***PSGrid::debug()***\n";
  cout<<code<<":P="<<P<<", refine="<<refine<<", delta="<<delta
      <<"\n"<<code<<"ORIGIN:\n"<<(*origin);
  cout<<code<<":GRID:\n"<<(*gridvect);
  cout<<code<<":******\n";
}/*end debug*/

double PSGrid::GetDelta() const {
/****
OUTPUT: returns delta
****/
  return delta;
}/*end GetDelta*/

bool PSGrid::IsFeasible(const Bounds &b, const mp_vector &x) const {
/****
INPUT: The bounds and an mp_vector.
OUTPUT: True if the point is within the founds, false if otherwise
****/                                             
  bool ret=true;
  for(long i=0;i<P&&ret;i++) {
    if ((b.lower[i]>x[i])||(b.upper[i]<x[i])) ret=false;
  }/*end for*/
  return ret;  
}/*end IsFeasible*/

long PSGrid::FindBestPSMultiple(const Bounds &b) {
/****
INPUT: The bounds of the space.
OUTPUT: Finds the best multiple for delta for use in the PatternSearch.  it
does so with the following methodology:
1) Find i s.t. i maximizes delta*grid[i] / b.upper[i]-b.lower[i].
2) Returns half the biggest multiple s.t.
mult * delta*grid[i] < b.upper[i]-b.lower[i].       
****/
  long i=0;
  double ratio=0;
  long mult=1;
  mp_vector d = b.upper - b.lower;

  /*Find the optimal i*/
  for(long j=0; j<P;j++) 
    if((*gridvect)[j][j] / d[j] > ratio) {i=j; ratio=(*gridvect)[j][j]/d[j];}

  ratio*=delta;

  /*Find the maximal multiple*/
  mult = (long) (d[i]/ratio);

  mult/=4; /*find the largest less than d[i]*/

  return mult;  
}/*end FindBestPSMultiple*/


/*******************************ACTION FUNCTIONS******************************/

void PSGrid::SnapToGrid(const Bounds &b, const mp_vector &old_pt,
                        mp_vector* new_pt) const {
/****
INPUT: The bounds and two mp_vectors - one reference and one pointer.
OUTPUT: The nearest grid point to old_pt will be returned in new_pt
NOTES: Its amazing what a little bit of linear algebra can do!  Note that this
will not necessarily work for grids that aren't parallel to the bounds.
****/

  ml_vector ipiv(P);
  pt_collect T(P,P);
  long i;/*counter*/
  mp_vector x=old_pt-(*origin);
#if DEBUG>=1
  if(gridvect==NULL) cerr<<"PSGSTG :Grid not initialized\n";
#endif

  T=(*gridvect)*delta;
  for(i=0;i<P;i++) x[i]=round(x[i]/T[i][i]);
  
  (*new_pt)= (*origin)+T*x;
  /*BOUNDS CHECKS*/
  for(i=0;i<P;i++) {
    if((*new_pt)[i]<b.lower[i]) (*new_pt)[i] = b.lower[i];
    if((*new_pt)[i]>b.upper[i]) (*new_pt)[i] = b.upper[i];  
  }/*end for*/

  
}/*end SnapToGrid*/


void PSGrid::RefineGrid(){
/****
EFFECT: Actually refines the grid. delta *= refine;
****/
  delta*=refine;
}/*end RefineGrid*/


void PSGrid::GetCore(const Bounds &b, const mp_vector &x,
                     pt_collect&core, long &pts) {
/****
INPUT: The bounds, a point, a pointer to the point collection to be returned,
and a reference passed number of points.
EFFECT: Returns the list of points in the Core Pattern of x, and the number of
points in this pattern. This should also call IsFeasible, and return only those
points within the bounds. Note that the returned values are stored on the rows
(aka each row represents a point in P-space), much as Eval_c is stored in
maps.h.
****/
  pt_collect temp(P,P);
  mp_vector test(P);
  pts=0;
  pt_collect t2(2*P,P);/*allocate more memory than we need*/
  long i,j; /*counters*/

  temp=(*gridvect)*delta;

  for(i=0;i<P;i++) {
    test=x+temp.col(i);/*create the potential core pattern vector*/
    if(IsFeasible(b,test)) {
      /*if feasible, copy it into a ROW of the outgoing matrix*/
      for(j=0;j<P;j++) t2[pts][j]=test[j];
      pts++;
    }/*end if*/

    test=x-temp.col(i);/*create the potential core pattern vector*/
    if(IsFeasible(b,test)) {
      /*if feasible, copy it into a ROW of the outgoing matrix*/
      for(j=0;j<P;j++) t2[pts][j]=test[j];
      pts++;
    }/*end if*/    
  }/*end for*/

  /*Now play the memory dance, and reclaim what was once ours!*/
  core.newsize(pts,P);
  for(i=0;i<pts;i++)
    for(j=0;j<P;j++)
      core[i][j]=t2[i][j];
}/*end GetCore*/
  
void PSGrid::SetDefault(long P1, const Bounds &b, const mp_vector &resolution){
/****
INPUT: Dimension, bounds and a grid resolution.
EFFECTS: This will automagically generate a simple grid that is parallel to
the axes.
****/
  long i;
  
#if DEBUG>=3
  cerr<<"PSGSD  :Called. resolution="<<resolution;
#endif  
  P=P1;
  gridvect=new pt_collect(P,P);
  origin=new mp_vector(b.lower);
  refine=PSG_DEFAULT_REFINE;
  delta=resolution[0];

  /*Finds smallest suggested resolution to make it delta*/
  for(i=0;i<P;i++)
    if(resolution[i]<delta)
      delta=resolution[i];
  
  for(i=0;i<P;i++)
    for(long j=0;j<P;j++)
      if(i!=j) (*gridvect)[i][j]=0;
      else (*gridvect)[i][j]=round(resolution[i]/delta);
  
#if DEBUG>=3
  cerr<<"PSGSD  :Done\n";
#endif
}/*end SetDefault*/



