/****
PSGrid class - 6/19/00 cmsief
This deals with the grid.  It assumes Compass Search for the Core Pattern.
In terms of the grid object, view it as if it were a real honest to god
matrix, unlike its counterpart in maps.h.

MAPS uses this class to find the core pattern around a point and to find
the best start step for the optimization.

NOTE: When choosing your grid or delta, it is probably a good idea that
you pick them s.t. delta*grid[i][i] divides evenly into the amount of
space in the ith dimension.  This allows the program to be guaranteed
to pick up on minima on the boundaries.
SetDefault should do this automatically.

ANOTHER NOTE: The implementation of SnapToGrid only deals with grids
parallel to the bounds.  You will have to reimplement it in order to
deal with other grids.
****/

#ifndef _MAPS_PSGRID_
#define _MAPS_PSGRID_
#include "maps_general.h"

#define PSG_DEFAULT_RESOLUTION 0.01
/*aka the default grid resolution is 1/100th of the space*/
#define PSG_DEFAULT_REFINE 0.5
/*aka halve delta by default*/


class PSGrid {
public:
  PSGrid();
  PSGrid(const PSGrid &g);
  PSGrid(long P1, const mp_vector& o, const pt_collect& g,
         double ref, double d);  

  PSGrid(long P1, const Bounds &b);
  /*if you aren't sure what to use, use this.*/
  
  PSGrid(long P1, const Bounds &b, const mp_vector &resolution);
  /*Resolution is in actual units, not percent of the space*/
  
  ~PSGrid();
  PSGrid & operator=(PSGrid &g);

  /*****RESET FUNCTIONS*****/
  
  void ResetPSGridState(long P1, const mp_vector& o,
                        const pt_collect& g, double ref,
                        double delta);
  /****
  INPUT: A complete set of state values for the grid.
  EFFECT: It copies these into the object's state.  As for your choice of delta
  and the grid, see the note above. 
  ****/
 
  bool ReadGridFromFile(char* fn);
  /****
  INPUT: File name string.
  OUTPUT: Returns true if read was sucessful.
  EFFECT: Reads the grid data from a file.
  FILE FORMAT: dimension<cr> 
               delta<cr>
               origin(P)<cr>
               refine(P)<cr>
               grid(PxP)
  NOTES: Watch VERY carefully what you do with directions,  When you write your
  grid to a file, assume that it is written like a real matrix. Also, keeping
  the grid vectors positive would be a good idea.  As for your choice of delta
  and the grid, see the note above.
  ****/

  /*****QUERY FUNCTIONS*****/

  void debug(char *code) const;
  /****
  EFFECT: does a complete state dump to stdout.
  ****/       

  double GetDelta() const;
  /****
  OUTPUT: returns delta.
  ****/
  
  bool IsFeasible(const Bounds &b, const mp_vector &x) const;
  /****
  INPUT: The bounds and an mp_vector.
  OUTPUT: True if the point is within the bounds, false if otherwise
  ****/  

  long FindBestPSMultiple(const Bounds &b);
  /****
  INPUT: The bounds of the space.
  OUTPUT: Finds the best multiple for delta for use in the PatternSearch.  it
  does so with the following methodology:
  1) Find i s.t. i maximizes delta*grid[i] / b.upper[i]-b.lower[i].
  2) Returns half the biggest multiple s.t.
  mult * delta*grid[i] < b.upper[i]-b.lower[i].       
  ****/
  
  /*****ACTION FUNCTIONS*****/
  
  void SnapToGrid(const Bounds &b, const mp_vector &old_pt,
                  mp_vector* new_pt) const;
  /****
  INPUT: The bounds and two mp_vectors - one reference and one pointer.
  OUTPUT: The nearest grid point to old_pt will be returned in new_pt
  ****/

  void GetCore(const Bounds &b, const mp_vector &x, pt_collect& core,
               long &pts);
  /****
  INPUT: The bounds, a point, a reference to point collection to be returned,
  and a reference passed number of points.
  EFFECT: Returns the list of points in the Core Pattern of x and the number
  of points in this pattern. This should also call IsFeasible, and return only
  those points within the bounds. Note that the returned values are stored on
  the rows (aka each row represents a point in P-space), much as Eval_c is
  stored in maps.h.
  ****/
  
  void RefineGrid();
  /****
  EFFECT: Actually refines the grid. delta *= refine;
  ****/
  
private:

  void SetDefault(long P1, const Bounds &b, const mp_vector &resolution);
  /****
  INPUT: Dimension, bounds and a grid resolution.
  EFFECT: This will automagically generate a simple grid that is parallel to
  the axes.  It will be similar to the default settings you give it, but not
  exactly that.
  ****/

  long P;               /*dimension of space*/
  mp_vector* origin;    /*The origin for the purpose of the grid (P)*/
  pt_collect* gridvect; /*The grid matrix.  This is in terms of multiples of
                          delta, hence it does not take scaling into account in
                          and of itself.*/
  double refine;        /*Grid refinement constant - aka what delta is
                          multiplied by when the grid is refined*/
  double delta;         /*delta from PatternSearch -
                          it is the current 'step size'*/
  
};/*end class*/

#endif


