/****
SearchCriterion Class - 6/27/00 cmsief
This is the search criterion class.
****/
#ifndef _MAPS_SC_
#define _MAPS_SC_

#include "maps_general.h"
#include "Approximate.h"
#include "NullApprox.h"
#include "KrigApprox.h"

/*weighting schemes*/
#define SC_WEIGHT_MULTIPLE 0
#define SC_WEIGHT_TARGET_VALUE 1
#define SC_WEIGHT_DELTA_MULTIPLE 2
#define SC_WEIGHT_DYNAMIC_GUESS 3
#define SC_WEIGHT_DYNAMIC 4 

class SearchCriterion {

public:
  SearchCriterion();
  SearchCriterion(long dim,Bounds &b);
  /*prep work and invokes SetDefault*/
  SearchCriterion(const SearchCriterion &S);
  SearchCriterion(Approximate &A, Bounds &b);
  ~SearchCriterion();
  SearchCriterion& operator=(SearchCriterion &S);
  
  /*****RESET FUNCTIONS*****/

  bool ResetSearchCriterionState(Approximate &A, int scheme_number, mp_vector &params);
  /****
  INPUT: An Approximation, a weight scheme number and its parameters.
  OUTPUT: returns true iff ResetWeightScheme returns true.
  EFFECT: Calls ResetApproximateion, and ResetWeightScheme.
  ****/
  
  void ResetApproximation(Approximate &A);
  /****
  INPUT: An Approximation.
  EFFECT: (*APX) = A;
  ****/
  
  bool ResetWeightScheme(int scheme_number, mp_vector &params);
  /****
  INPUT: The scheme number and its parameters.
  OUTPUT: Returns true if everything is valid.
  EFFECT: This sets the w_scheme integer, and then sets w_params.  This will
  also set w_c to the correct value, based on our current evaluation number.
  ****/

  /*****QUERY FUNCTIONS*****/

  void debug(char *code) const;
  /****
  EFFECT: does a complete state dump to stdout.
  ****/ 
  
  /*****ACTION FUNCTIONS*****/
  
  void EvaluateSearchCriterion(long dim, mp_vector &x, double &f, bool &success);
  /****
  INPUT: The dimension of the space, a list of values, a reference passed value
  of the function, and a reference-passed flag.  The input syntax is this
  convoluted to ensure that it works with Anne Shepherd's version of Liz Dolan's
  PatternSearch.
  EFFECT: This evaluates the search criterion at a given point, calling
  Approximate::Evaluate to do so, and takes the weighting schedule into
  account.  If the evaluation was a success, then success is 1, else it is
  set to 0.
  ****/

  bool UpdateSearchCriterion(pt_collect &points, mp_vector& values, long
                             numpoints, long n, double delta);
  /****
  INPUT: The list of evaluated sites, the values, and the number of points
  listed, the number of points in the initial design, and the curent delta.
  OUTPUT: Returns true if the operation was sucessful.
  EFFECT: This will call Approximate::UpdateApprox to update the approximation, then
  will update the weighting scheme by calling the appropriate UpdateWeight function.
  ****/

  friend void OptimizeSearchCriterion(long dimx, mp_vector &x, double &f, bool &success, void * SC) 
    {(*(SearchCriterion*)SC).EvaluateSearchCriterion(dimx,x,f,success);}
  /****
  INPUT: PatternSearch calling scheme.  See PatternSearch.h
  EFFECT: Indirection to get around C++'s inability to let function pointer
  point to class member functions.  This gets optimized by PatternSearch.
  ****/

  void GetApproximationGraphics(long dim, mp_vector &x, double &f, bool &success);
  /***
  INPUT: PatternSearch calling scheme.  See PatternSearch.h
  OUTPUT: Value of the Approximation only at that point.  Meant for use by
  GenerateMAPSGraphics
  ****/
  
private:  

  void SetDefault(long dim);
  /****
  INPUT: Dimension of space.
  EFFECT: Automagically generates one of these objects, uses (2,1.25,0.75)
  WeightDynamic scheme and Kriging Approximation.
  ****/
  
  /*Weighting Scheme functions*/

  void CreateWeightTargetValue(mp_vector &params);
  /*This does the computation of w_params[3]
    m = (p[1] / p[0])^(1/p[2])*/
  void UpdateWeightTargetValue(long c);
  /****
  INPUT: The iteration number.
  EFFECT: Multiply w_c by the correct multiplier in order to reach the level of
  insignificance by the specified iteration.
  PARAMS FORMAT: [0] - w_0
                 [1] - target value of insignificance
                 [2] - iteration number to reach insignificance
  Though w_params[3] is not specified by the user, when you install the Target
  Value weighting scheme, the multiple is computed and stored here.
  ****/
  
  void UpdateWeightMultiple(long c);
  /****
  INPUT: The iteration number.
  EFFECT: This will multiply w_c by w_params[1]
  PARAMS FORMAT: [0] - w_0
                 [1] - m, s.t. w_c = m * w_{c-1}
  ****/

  void UpdateWeightDeltaMultiple(double curr_delta);
  /****
  INPUT: current delta
  EFFECT: If delta has been refined, the weight is multiplied by w_1
  PARAMS FORMAT: [0] - w_0, the weight before delta refinement
                 [1] - m, s.t. w_c = m*w_{c-1}, if delta was refined last iteration
                                   = w_{c-1}, if delta wasn't refined last iteration
  ****/

  void UpdateWeightDynamicGuess(long lastsite, mp_vector& values);
  /****
  INPUT: number of last site eval'd and vector of function values
  EFFECT: If the approximation was 'close' to correct at the last site, then we
  reduce the weight, less we increase it.
  PARAMS FORMAT: [0] - w_0
                 [1] - t, the threshhold percentage
                 [2] - m_b, multiplier if approximation is 'bad'
                 [3] - m_g, multiplier if approximation is 'good'
                 
                 w_c =  m_g * w_{c-1}, if |f(x_{c-1}) -\hat{f_{c-1}(x_c-1)}| <= t*frange
                     =  m_b * w_{c-1}, otherwise
  ****/

  void UpdateWeightDynamic(long lastsite, mp_vector& values);
  /****
  INPUT: number of last site eval'd and vector of function values
  EFFECT: If the approximation was 'close' to correct at the last site, then we
  reduce the weight, less we increase it.  Similar to DynamicGuess, except it uses 
  |expval - bestpoint| / 2 as the threshold. 
  PARAMS FORMAT: [0] - w_0
                 [1] - m_b, multiplier if approximation is 'bad'
                 [2] - m_g, multiplier if approximation is 'good'
                 
                 w_c =  m_g * w_{c-1}, if |f(x_{c-1}) -\hat{f_{c-1}(x_c-1)}| <= |expval-best|/2
                     =  m_b * w_{c-1}, otherwise
  ****/
  
  
  Bounds *bds;             /*the bounds*/
  //long P;                  /*dimension of space*/
  int w_scheme;            /*Weighting scheme number*/
  mp_vector* w_params;     /*Parameters for that scheme.  In general w_params[0]
                            will always be w_0*/
  long np;                 /*The number of points evaluated so far*/  
  Approximate* APX;        /*Some kind of Approximation*/ 
  pt_collect* Eval_c;      /*A pointer to the evaluated points list*/
  mp_vector* f_values;      /*A pointer to the function vals*/
  double w_c;              /*w_c for the current iteration*/
  double original_delta;   /*original delta - for Delta Based Weight Scheme*/
  bool first_time;         /*to prevent weight recomputation for the first time*/
};/*end class*/

#endif
