/*MAPS IMPLEMENTATION FILE - 8/1/00*/

#include <cstdio>
#include <iostream>
#include <fstream>
#include <unistd.h>
#include <cstdlib>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <malloc.h>
#include <cstring>
#include "maps.h"

/*Iteration History Field Widths*/
#define _GFN 4
#define _GFP 10
#define _GFV 10

/*Official stopping tolerance for MAPS*/
double TOLERANCE=10e-12;

using namespace std;

maps::maps() {
/*status=NOTINIT*/
  V=N=C=P=x_c=0;
  status=NOTINIT;g=NULL;
  Eval_c=NULL;f_values=NULL;Fname=NULL;
  ID=NULL;SC=NULL;PS=NULL;initial_guess=NULL;
  TOLERANCE=10e-12;
}/*end default constructor*/
  
maps::maps(long V1, long N1, long P1, const Bounds &B1, char* fn,
           const PSGrid &G1, const InitialDesign &I,
           const SearchCriterion &S1, const PatternSearch &PS1) {
/*Prep work and call ResetMAPSState*/
  ID=NULL; SC=NULL; PS=NULL;g=NULL;initial_guess=NULL;TOLERANCE=10e-12;
  ResetMAPSState(V1,N1,P1,B1,fn,G1,I,S1,PS1);
}/*end special constructor*/
  
maps::maps(long V1, long P1, Bounds &B1, char*fn) {
  /*Prep work and call SetDefault*/
  ID=NULL; SC=NULL; PS=NULL;g=NULL;initial_guess=NULL;TOLERANCE=10e-12;
  SetDefault(V1,P1,B1,fn);
}/*end special constructor*/

maps::maps(maps& old){
  ID=NULL;SC=NULL;PS=NULL;g=NULL;
  bool success=true;
  V=old.V;N=old.N;P=old.P;b=old.b;x_c=old.x_c;
  status=READY;/*just temporary*/
  Fname=NULL;
  if(old.Fname!=NULL) success=ResetFunction(old.Fname);
  g=NULL;
  if(old.g!=NULL) success=success&&ResetGrid(*old.g);
  ID=NULL;
  if(old.ID!=NULL) ID = new InitialDesign(*old.ID);
  SC=NULL;
  if(old.SC!=NULL) success=success&&ResetSearchCriterion(*old.SC);
  PS=NULL;
  if(old.PS!=NULL) success=success&&ResetPatternSearch(*old.PS);
  if(old.Eval_c!=NULL) Eval_c = new pt_collect(*old.Eval_c);
  else Eval_c=NULL;
  if(old.f_values!=NULL) f_values = new mp_vector(*old.f_values);
  else f_values=NULL;
  if(old.initial_guess!=NULL) initial_guess =
                                  new mp_vector(*old.initial_guess);
  else initial_guess=NULL;
  status=old.status;
#if DEBUG>=1
  if(!success) cerr<<"MPCC   : One items in the copy failed.\n";
#endif
}/*end Copy Constructor*/

maps::~maps() {
  if(ID!=NULL) delete ID;
  if(SC!=NULL) delete SC;
  if(PS!=NULL) delete PS;
  if(Eval_c!=NULL) delete Eval_c;
  if(f_values!=NULL) delete f_values;
  if(Fname!=NULL) delete Fname;
  if(g!=NULL) delete g;
  if(initial_guess!=NULL) delete initial_guess;
}/*end destructor*/

/*****RESET FUNCTIONS****/

bool maps::ResetMAPSState(long V1, long N1, long P1, const Bounds &B1,
                          char* fn, const PSGrid &G1,
                          const InitialDesign &I1, const SearchCriterion &S1,
                          const PatternSearch &PS1) {
/****
INPUT: A partial set of state parameters for MAPS
OUTPUT: Returns true if the operation was successful.
EFFECT: Sets all the parameters, and sets status=READY,
clears out Eval_c and f_values.  This wipes out everything else entirely.
*****/
  bool success;
  NewMAPSRun();/*wipes out Eval_c and f_values*/
  V=V1;N=N1;P=P1;b=B1;
  success=ResetFunction(fn)&&ResetGrid(G1)&&ResetDesign(I1)&&
    ResetSearchCriterion(S1)&&ResetPatternSearch(PS1);
#if DEBUG>=1
  if(!success) cerr<<"MPRMS  : One or more resets failed.\n";
#endif
  return success;
}/*end ResetMAPSState*/
  
void maps::NewMAPSRun(){
/****
EFFECT: Clears the Eval_c and F_values,
resets C=0 and sets status=READY.
This allows you to reset MAPS between runs on a given function.
****/ 
  if(Eval_c!=NULL) {delete Eval_c;Eval_c=NULL;}
  if(f_values!=NULL) {delete f_values;f_values=NULL;}
  C=0;x_c=0;
  status=READY;  
}/*end NewMAPSRun*/


bool maps::NewInitialPoint(const mp_vector& newpoint){
/****
INPUT: An initial guess for the starting point
OUTPUT: True if the initial point was set. False on error.
EFFECT: If the point is of the right dimension and in bounds,
it will be used and true will be returned. 
****/
  bool success=true;

  /*Dim check*/
  if(newpoint.dim()!=b.lower.dim()) return false;

  /*Bounds check*/
  for (long i=0;success && i<newpoint.dim();i++)
    if(newpoint[i]<b.lower[i]||newpoint[i]>b.upper[i])
      success=false;
    
  if(initial_guess==NULL) initial_guess=new mp_vector(1);
  (*initial_guess)=newpoint;
  
  return success;  
}/*end NewInitialPoint*/

  
bool maps::ResetGrid(const PSGrid &G1) {
/****
INPUT: A PSGrid
OUTPUT: If it copies G1 it returns true, else it returns false.
EFFECT: If status==READY, it copies G1 to G, else no effect.  
****/
  if(status==READY) {
    if(g!=NULL) delete g;
     g = new PSGrid(G1);
    return true;
  }/*end if*/
  else return false;
}/*end ResetGrid*/
  
bool maps::ResetDesign(const InitialDesign &I) {
/****
INPUT: An InitialDesign.
OUTPUT: If it copies I it returns true, else it returns false.
EFFECT: If status==READY and I.IsDesignInitialized()==true,
it copies I to ID, else no effect.  
****/
 if(status==READY&&I.IsDesignInitialized()) {
    if(ID!=NULL) delete ID;
    ID = new InitialDesign;
    (*ID) =I;
    N=(*ID).GetDesignSize();
    return true;
  }/*end if*/
  else return false;
}/*end ResetDesign*/
  
bool maps::ResetSearchCriterion(const SearchCriterion &S){
/****
INPUT: A SearchCriterion.
OUTPUT: If it copies S it returns true, else it returns false.
EFFECT: If status==READY, it copies S to SC, else no effect.
****/
  if(status==READY) {
    if(SC!=NULL) delete SC;
    SC = new SearchCriterion(S);
    return true;
  }/*end if*/
  else return false;
}/*end ResetSearchCriterion*/
  
bool maps::ResetPatternSearch(const PatternSearch &PS1) {
/****
INPUT: A PatternSearch object.
OUTPUT: If it copies S it returns true, else it returns false.
EFFECT: If status==READY, it copies PS1 to PS, else no effect.
****/
  if(status==READY) {
    if(PS!=NULL) delete PS;
    mp_vector temp=0;
    PS = new CompassSearch(P,temp);
    (*PS) = PS1;
    return true;
  }/*end if*/
  else return false;
}/*end ResetPatternSearch*/
  
bool maps::ResetFunction(char* fn){
/****
INPUT: String containing the name of the function to be optimized.
OUTPUT: Returns true.
EFFECT: Fname gets fn. 
****/
  if(Fname!=NULL) free(Fname);
  Fname=(char*)malloc(strlen(fn)+1);
  strcpy(Fname,fn);
  return true;
}/*end ResetFunction*/
  
  /*****QUERY FUNCTIONS*****/
  
bool maps::RetrieveBestSeen(mp_vector* bpoint, double* bvalue) const{
/****
INPUT: Two Pointers, one for a mp_vector, one for a value
OUTPUT: True if status==DONE, false if otherwise.
EFFECT: If status=DONE, it finds the best point and its value and
assigns it to the pointers.  This is done using a linear search.
****/
  long best;
  double bval;
  
  if(status==MAPS_DONE) {
    best=0;bval=(*f_values)[0];
    /* Go through evaluated points and get minimum */
    for(long i=0;i<V;i++)
      if((*f_values)[i]<bval) {
        bval=(*f_values)[i];
        best=i;        
      }/*end if*/
    
    (*bpoint)=(*Eval_c).row(best);
    (*bvalue)=bval;
    return true;
  }/*end if*/
  else return false;
}/*end RetrieveBestSeen*/
  
void maps::debug(char* code) const{
/****
EFFECT: Does a complete state dump to stdout.
This should do state dumps of all components.
*****/       
  cout<<code<<":***maps::debug()***\n";
  cout<<code<<":V="<<V<<" N="<<N<<" C="<<C<<" P="<<P<<" x_c="<<x_c<<endl;
  cout<<code<<":status=";
  switch(status) {
  case NOTINIT:
    cout<<"NOTINIT";break;
  case READY:
    cout<<"READY";break;
  case MAPS_DONE:
    cout<<"MAPS_DONE";break;
  default:
    cout<<"ERROR!";
  };/*end switch*/
  cout<<" Fname="<<Fname<<endl;
  
  cout<<code<<":b.lower="<<b.lower;
  cout<<code<<":b.upper="<<b.upper;
  if(Eval_c!=NULL) cout<<code<<":Eval_c="<<(*Eval_c);
  else cout<<code<<":Eval_c=NULL!\n";
  if(f_values!=NULL) cout<<code<<":f_values="<<(*f_values);
  else cout<<code<<":f_values=NULL!\n";
  if(g!=NULL) (*g).debug(code);
  else cout<<code<<":g=NULL!\n";
  if(SC!=NULL) (*SC).debug(code);   
  else cout<<code<<":SC=NULL!\n";
  cout<<code<<":***end MP:debug()***\n";   
}/*end debug*/
  
void maps::GenerateMAPSGraphics(char* fpref) const{
/****
INPUT: Filename prefix.
EFFECT: Outputs to fpref.MAPS.dat stuff that would be nice in
producing pictures via GNUplot.  Due to limits in visualization
technology, this feature is only available for functions from R or
R^2.

MAPS Manual, Section B explains how to generate graphics with gnuplot
****/

  if(status!=MAPS_DONE && status!=MAPS_RUNNING) return;
  
  if(status==MAPS_DONE) {
    (*SC).UpdateSearchCriterion(*Eval_c,*f_values,V,N,(*g).GetDelta());
  }/*end if*/
  /*this is to update the SC for the last point.*/
  
  /* One-dimensional function */
  if(P==1) {
    ofstream ofs_sc,ofs_ap,ofs_gf;
    char* fn_sc = (char*)malloc(strlen(fpref) +13);
    char* fn_ap = (char*)malloc(strlen(fpref) +13);
    char* fn_gf = (char*)malloc(strlen(fpref) +14);
    mp_vector y(P);
    double f;
    bool s;
    mp_vector diff=(b.upper-b.lower) * (1.0/500.0);
    
    sprintf(fn_sc,"%s.MAPS.SC.dat",fpref);
    sprintf(fn_ap,"%s.MAPS.AP.dat",fpref);
    sprintf(fn_gf,"%s.MAPS.PTS.dat",fpref);
    ofs_sc.open(fn_sc,ios::out);
    ofs_ap.open(fn_ap,ios::out);
    ofs_gf.open(fn_gf,ios::out);
    ofs_sc<<"# This Graphics File has been brought to you by MAPS\n"
          <<"# implemented by: Chris Siefert, College of William and Mary\n"
          <<"# for parameters V="<<V<<" N="<<N<<" P="<<P;
    if(g!=NULL) ofs_sc<<" delta="<<(*g).GetDelta();
    ofs_sc<<endl
          <<"# on function: "<<Fname<<endl;
    ofs_sc<<"# Current Search Criterion Graph:\n";
    ofs_ap<<"# This Graphics File has been brought to you by MAPS\n"
          <<"# implemented by: Chris Siefert, College of William and Mary\n"
          <<"# for parameters V="<<V<<" N="<<N<<" P="<<P;
    if(g!=NULL) ofs_ap<<" delta="<<(*g).GetDelta();
    ofs_ap<<endl    
          <<"# on function: "<<Fname<<endl;
    ofs_ap<<"# Current Approximation Graph:\n";
    ofs_gf<<"# This Graphics File has been brought to you by MAPS\n"
          <<"# implemented by: Chris Siefert, College of William and Mary\n"
          <<"# for parameters V="<<V<<" N="<<N<<" P="<<P;
    if(g!=NULL) ofs_gf<<" delta="<<(*g).GetDelta();
    ofs_gf<<endl
          <<"# on function: "<<Fname<<endl;
    ofs_gf<<"# Current Evaluated sites:\n";

    for(long m=0;m<C;m++)
      ofs_gf<<setprecision(4)<<(*Eval_c)[m][0]<<" "<<(*f_values)[m]<<endl;    
    
    for(double k=b.lower[0];k<b.upper[0];k+=diff[0]) {    
        y[0]=k; 
        (*SC).EvaluateSearchCriterion(P,y,f,s);
        ofs_sc<<setprecision(4)<<y[0]<<" "<<setprecision(8)<<f<<"\n";
        (*SC).GetApproximationGraphics(P,y,f,s);
        ofs_ap<<setprecision(4)<<y[0]<<" "<<setprecision(8)<<f<<"\n";
      }/*end for*/
    
    free(fn_sc);free(fn_ap);free(fn_gf);
    ofs_gf.close();ofs_sc.close();ofs_ap.close();   
  }/*end if P==1*/
  /* Two-dimensional function */
  else if(P==2) {
    ofstream ofs_sc,ofs_ap, ofs_gf;
    char* fn_sc = (char*)malloc(strlen(fpref) +13);
    char* fn_ap = (char*)malloc(strlen(fpref) +13);
    char* fn_gf = (char*)malloc(strlen(fpref) +14);
    mp_vector y(P);
    double f;
    bool s;
    mp_vector diff=(b.upper-b.lower) * (1.0/50.0);
    
    sprintf(fn_sc,"%s.MAPS.SC.dat",fpref);
    sprintf(fn_ap,"%s.MAPS.AP.dat",fpref);  
    sprintf(fn_gf,"%s.MAPS.PTS.dat",fpref);   
    ofs_sc.open(fn_sc,ios::out);
    ofs_ap.open(fn_ap,ios::out);
    ofs_gf.open(fn_gf,ios::out);
    
    ofs_sc<<"# This Graphics File has been brought to you by MAPS\n"
          <<"# implemented by: Chris Siefert, College of William and Mary\n"
          <<"# for parameters V="<<V<<" N="<<N<<" P="<<P;
    if(g!=NULL) ofs_sc<<" delta="<<(*g).GetDelta();
    ofs_sc<<endl
          <<"# on function: "<<Fname<<endl;
    ofs_sc<<"# Current Search Criterion Graph:\n";
    ofs_ap<<"# This Graphics File has been brought to you by MAPS\n"
          <<"# implemented by: Chris Siefert, College of William and Mary\n"
          <<"# for parameters V="<<V<<" N="<<N<<" P="<<P;
    if(g!=NULL) ofs_ap<<" delta="<<(*g).GetDelta();
    ofs_ap<<endl
          <<"# on function: "<<Fname<<endl;
    ofs_ap<<"# Current Approximation Graph:\n";

    ofs_gf<<"# This Graphics File has been brought to you by MAPS\n"
          <<"# implemented by: Chris Siefert, College of William and Mary\n"
          <<"# for parameters V="<<V<<" N="<<N<<" P="<<P;
    if(g!=NULL) ofs_gf<<" delta="<<(*g).GetDelta();
    ofs_gf<<endl
          <<"# on function: "<<Fname<<endl;
    ofs_gf<<"# Current Evaluated sites:\n";

    for(long m=0;m<C;m++)
      ofs_gf<<setprecision(4)<<(*Eval_c)[m][0]<<" "
            <<(*Eval_c)[m][1]<<" "<<(*f_values)[m]<<endl;   
    
    for(double i=b.lower[0];i<b.upper[0];i+=diff[0]) {
      y[0]=i;
      for(double j=b.lower[1];j<b.upper[1];j+=diff[1]) {    
        y[1]=j; 
        (*SC).EvaluateSearchCriterion(P,y,f,s);
        ofs_sc<<setprecision(4)<<y[0]<<" "<<y[1]<<" "
              <<setprecision(8)<<f<<"\n";
        (*SC).GetApproximationGraphics(P,y,f,s);
        ofs_ap<<setprecision(4)<<y[0]<<" "<<y[1]<<" "
              <<setprecision(8)<<f<<"\n";
      }/*end for*/
      ofs_sc<<"\n";
      ofs_ap<<"\n";
    }/*end for*/

    free(fn_sc);free(fn_ap);free(fn_gf);
    ofs_sc.close();ofs_ap.close();ofs_gf.close();
  }/*end if P==2*/
  else return;
}/*end GenerateMAPSGraphics*/

void maps::GenerateIterationHistory() const{
/****
EFFECT: Outputs the chosen points, their values and the best point and values
returned to stdout.  To convert these data files into LaTeX tables with ease,
use the included mdat2tex.awk script with the following syntax :
awk -f mdat2tex.awk <mapsdatafile>
awk will output the LaTeX to stdout.
****/  
  if(status!=MAPS_DONE) return;
  
  cout<<"******MAPS ITERATION HISTORY******\n"
      <<"# for parameters V="<<V<<" N="<<N<<" P="<<P;
  if(g!=NULL) cout<<" delta="<<(*g).GetDelta();
  cout<<endl
      <<"# on function: "<<Fname<<endl;

  /* Table headers */
  cout<<setw(_GFN)<<"NUM ";
  for(long k=0;k<P;k++) cout<<setw(_GFP)<<k+1<<" ";
       /*+1 notation as per Trosset*/
  cout<<setw(_GFV)<<"VAL\n";
  
  /* Print x and f for all evaluations */
  for(long i=0;i<V;i++) {
    cout<<setw(_GFN)<<i+1<<" "; /*+1 notation as per Trosset*/
    for(long j=0;j<P;j++) cout<<setw(_GFP)<<setprecision(_GFP-3)
                              <<(*Eval_c)[i][j]<<" ";
    cout<<setw(_GFV)<<setprecision(_GFV-2)<<(*f_values)[i]<<endl;
  }/*end if*/

  /* Print Best Point */
  cout<<"\nBest Point found is #"<<x_c+1<<", value="<<setw(_GFV)
      <<(*f_values)[x_c]<<endl;
  /*+1 notation as per Trosset*/
  cout<<"**********************************\n";

}/*end GenerateIterationHistory*/

  /*****ACTION FUNCTIONS*****/
  
bool maps::RunMAPS(char * fpref){
/****
INPUT: a pointer that specifies the filename prefix for
iteration-by-iteration graphics.  Set this to NULL for no graphics,
or call the parameter-free version RunMAPS()
OUTPUT: Returns true if nothing crashed.
EFFECT: Assume status=READY.  Then it evaluates the function at the initial
design sites. Then it begins the iterative process of looping until we use up
the evaluation budget. When it finishes, it sets status=MAPS_DONE.  This will
call EvalF.  If fpref!=NULL, then this will output graphics to files named :
<fpref><C>.MAPS.AP.dat and <fpref><C>.MAPS.SC.dat .
****/       
  
  /*Sanity checks*/
  if(status!=READY||g==NULL||ID==NULL||SC==NULL||PS==NULL) {
#if DEBUG>=1
    cerr<<"MPRM   :MAPS not initialized correctly\n";
#endif
    return false;
  }/*end if*/

  if(Eval_c!=NULL) delete Eval_c;
  if(f_values!=NULL) delete f_values;
  status=MAPS_RUNNING;
  
  /*Variable Declarations*/
  pt_collect temp_eval(V,P);/*a size VxP array, for resizing Eval_c*/
  pt_collect temp_core(1,P);/*an array for the core pattern...
                              this will be resized*/  
  mp_vector prosp_pt(P);    /*a prospective point*/
  mp_vector prosp_pt2(P);   /*a prospective point...
                              for the SnapToGrid Checks*/   
  mp_vector temp_fval(V);   /*a size V array for resizing f_values*/
  long temp_core_pts;       /*number of points in the core pattern*/
  double temp_f;            /*value from a function evaluation*/
  bool success;             /*you just always need a bool:) */
  bool s;                   /*for calling EvaluateSearchCriterion*/
  long i,j;                 /*loop counters to the stars*/
  char *gfx_fp=NULL;        /*augmented fname for graphics*/
  bool converged=false;     /*stopping due to convergence*/

  if(fpref!=NULL) gfx_fp=(char*)malloc(1*sizeof(char));//just for 1st iteration


  /*Sanity Check - design size*/
  if(N!=(*ID).GetDesignSize()) {
#if DEBUG>=1
    cerr<<"MPRM   :WARNING - Initial Design Size(N) "
        <<"set incorrectly, resetting N...\n";
#endif
    N=(*ID).GetDesignSize();
  }/*end if*/
  
  /*Allocate space in Eval_c and f_values for N initial points*/
  /*Must be resized later*/
  Eval_c=new pt_collect(N,P);
  f_values=new mp_vector(N);

  /*Get initial design sites generated by Latin Hypercube (or other method)
    and put into Eval_c*/
  (*ID).GetDesign(Eval_c);
#if DEBUG>=2
  cout<<"MPRM   :Before evals...Eval_c="<<(*Eval_c);  
#endif

  /*Evaluate the function at the initial design sites*/
  for(i=0;i<N;i++) {
      /* Call the real function */
      temp_f=EvalF((*Eval_c).row(i),success);

      /* If function call was successful, set f_values.
         Also setting temp variables (Needed so we can
         later use newsize function on Eval_c and f_values) */
      if(success) {
          (*f_values)[i]=temp_f;
          temp_fval[i]=temp_f;
          for(j=0;j<P;j++) temp_eval[i][j] = (*Eval_c)[i][j];        
      }/*end if*/
    else {
#if DEBUG>=1
      cerr<<"MPRM   :Evaluation failed at point="<<(*Eval_c).row(i);
#endif
      //NEED TO DEAL WITH EVALUATION FAILURE!!!!!!!!!!!!!!!!!!
      abort();
    }/*end else*/    
  }/*end for*/

  /*Find point with minimal value and make it x_c*/
  x_c=0;
  temp_f=(*f_values)[0];
  for(i=0;i<N;i++)  
    if((*f_values)[i]<temp_f) {
      temp_f=(*f_values)[i];
      x_c=i;
    }/*end if*/

/* Print out extra information if ITERATE_DEBUG */
#ifdef ITERATE_DEBUG
  for(long w=0;w<N;w++) {
    cout<<w+1;
    if(w==x_c) cout<<"* ";
    else cout<<"  ";
    for(long r=0;r<P;r++) cout<<setprecision(10)<<(*Eval_c)[w][r]<<" ";
    cout<<(*f_values)[w]<<" delta="<<(*g).GetDelta()<<endl;
  }/*end for*/
#endif
  
  /*********************************************************************/
  /*ENTER THE MAIN LOOP - Iterate until converged or budget is expended*/
  for(C=N;!converged&&C<V;C++) {

    /*Main block*/
#if DEBUG>=2
    cout<<"MPRM    :******BEGIN ITERATION C="<<C<<"******\n";
    cout<<"MPRM    :x_c="<<x_c<<endl;
    cout<<"MPRM    :Eval_c="<<(*Eval_c);
    cout<<"MPRM    :Best Value ="<<(*f_values)[x_c]<<endl;
    cout<<"MPRM    :Best Point ="<<(*Eval_c).row(x_c);
    (*g).debug("MPRM    ");
#endif

    /*Get the Core Pattern around the current minimum (x_c) */
    (*g).GetCore(b,(*Eval_c).row(x_c), temp_core, temp_core_pts);

    /*if all pts in core have been evaluated, refine the grid*/
    if(PCSubset(temp_core, *Eval_c)) {
      (*g).RefineGrid();
#if DEBUG>=2
      cout<<"MPRM   :Grid Refined to delta="<<(*g).GetDelta()<<endl;
#endif      
      if((*g).GetDelta() < TOLERANCE) {
        /* Iteration loop will complete after one more pass */
        converged=true;
        break;
      }/*end if*/
    }/*end if*/
       
    /*Update the Search Criterion*/
    /*Eval_c must have dimensions (C,P) and f_values must be of
      size C, so they will have to be resized at every iteration*/
    success=(*SC).UpdateSearchCriterion(*Eval_c,*f_values,C,N,(*g).GetDelta());
#if DEBUG>=2
    cout<<"MPRM  :Eval_c="<<*Eval_c;
    cout<<"MPRM  :f_values="<<*f_values;
    cout<<"MPRM  :Search Criterion Updated\n";
#endif
    
#if DEBUG>=1
    if(!success) cerr<<"MPRM   :Search Criterion Update FAILED!\n";
#endif

    /*Iteration-by-Iteration Graphics code*/
    if(fpref!=NULL) {
      i=C+1;j=1;
      while(i>=10) {i/=10;j++;} /*count # of digits in C*/
      realloc(gfx_fp,strlen(fpref) + j + 1);
      sprintf(gfx_fp,"%s%ld",fpref,C+1);
      GenerateMAPSGraphics(gfx_fp);
    }/*end if*/
    
    /*We might have to pick a better starting point than *Eval_c.row(x_c).  If
      so, that will be done here.*/
    prosp_pt=(*Eval_c).row(x_c);    
   
    if(initial_guess!=NULL && C==N) {
      /*If the user has supplied an initial guess by
        calling NewInitialPoint, this is where we want
        to use it.*/
      prosp_pt=(*initial_guess);
    }/*end if*/
    else {
      /*Call the "oracle" to search and find the next prospective point*/
      PS->CleanSlate(P,prosp_pt, g->FindBestPSMultiple(b)* g->GetDelta(),
                     g->GetDelta(), OptimizeSearchCriterion, (void*)SC);
      /*Run the PS*/
      PS->BeginSearch();
      PS->GetMinPoint(prosp_pt);
    }/*end else*/

    /*MAKE SURE THE POINT IS ON THE GRID (WHICH IT SHOULD BE,
      BUT IT NEVER HURTS TO CHECK) AND MOVE IT IF IT'S NOT*/
    (*g).SnapToGrid(b,prosp_pt,&prosp_pt2);

    if(prosp_pt!=prosp_pt2) {
#if DEBUG>=2
      cout<<"MPRM   :ERROR(non-fatal) Pattern Search Point not on Grid:"<<prosp_pt
          <<"MPRM   :Replaced with:"<<prosp_pt2;
#endif
      prosp_pt=prosp_pt2;
    }/*end if*/

    /*Check to see if the prospective point has already been evaluated*/
    if(IsElement(prosp_pt, *Eval_c)) {
      /*this site has already been evaluated*/
      double best_core_val=ERROR;     /*Best S_c value in the core*/
      long best_core_num=ERROR;       /*Best S_c point in the core*/
#if DEBUG>=2
      cout<<"MPRM   :Site Already Eval'd, site="<<prosp_pt;
#endif
      
      /*Get the Core Pattern of the prospective point*/
      (*g).GetCore(b,prosp_pt, temp_core, temp_core_pts);
      
      /*Try to find a point in the core that hasn't been evaluated
        and is predicted to be the minimum in the core*/
      for(i=0;i<temp_core_pts;i++) {
        mp_vector another_temp=temp_core.row(i);
        if(!IsElement(temp_core.row(i),*Eval_c)) {          
          (*SC).EvaluateSearchCriterion(P,another_temp,temp_f, s);
          if(best_core_num==ERROR || temp_f<best_core_val){
            best_core_val=temp_f;
            best_core_num=i;
          }/*end if*/
        }/*end if*/
      }/*end for*/

      if(best_core_num==ERROR) {
          /* All points in the core of the prospective point
             have been evaluated. Fall back on Pattern Search
             around the current minimum (x_c) */
#if DEBUG>=1
          cerr<<"MPRM   :No valid point found in core.\n";//Will SegFault\n";
          cerr<<"MPRM   :Reval'd point="<<prosp_pt;
#endif	  
	  /*Get the Core Pattern of x_c again*/
	  (*g).GetCore(b,(*Eval_c).row(x_c), temp_core, temp_core_pts);

          /*Try to find a point in the core that hasn't been evaluated*/
	  success=false;
	  for(i=0;!success&&i<temp_core_pts;i++) {
	    if(!IsElement(temp_core.row(i),(*Eval_c)))
	      success=true;prosp_pt=temp_core.row(i);
	      }/*end if*/
#if DEBUG>=2
	  if(success)
              /* A new prospective point was found */
              cerr<<"MPRM   :Using alternate point="<<prosp_pt;
	  else{ 
	    abort();
	    cerr<<"MPRM   :No available re-eval pt.  Dumping core.\n";
	  }/*end else*/
#endif
      }/*end if*/
      else {
          /* Evaluate the best candidate in the prospective point core */
          prosp_pt=temp_core.row(best_core_num);      
      }/*end if*/
    } /*end if(IsElement(prosp_pt, *Eval_c))*/

#if DEBUG>=2
    cout<<"MPRM   :Prospective Point chosen, point="<<prosp_pt;
#endif
    
    /*Evaluate the Function*/
    temp_f=EvalF(prosp_pt,success);
    
    if(success) {
        /*Update temps before resize of Eval_c and f_values*/
        temp_fval[C] = temp_f;
        for(i=0;i<P;i++) temp_eval[C][i] = prosp_pt[i];
        
        /*Resize Eval_c and f_values*/
        (*Eval_c).newsize(C+1,P);
        (*f_values).newsize(C+1);
      
        /*Fill back in f_values and Eval_c after resize*/
        for(i=0;i<C+1;i++) {
            (*f_values)[i]=temp_fval[i];
            for(j=0;j<P;j++) (*Eval_c)[i][j] = temp_eval[i][j];        
        }/*end for*/
    }/*end if*/
    else {
#if DEBUG>=1
      cerr<<"MPRM   :Evaluation failed at point="<<prosp_pt;
#endif 
      //NEED TO DEAL WITH EVALUATION FAILURE!!!!!!!!!!!!!!!!!!      
      abort();
    }/*end else*/ 

    /*Update the current minimum (x_c) if needed*/
    if(temp_f < (*f_values)[x_c]) x_c=C;      

#ifdef ITERATE_DEBUG
      cout<<C+1;
      if(C==x_c) cout<<"* ";
      else cout<<"  ";
      for(long q=0;q<P;q++) cout<<setprecision(15)<<(*Eval_c)[C][q]<<" ";
      cout<<(*f_values)[C]<<" delta="<<(*g).GetDelta()<<endl;
#endif
    
  }/*end the main for loop*/

#if DEBUG>=1
  if(converged) cout<<"MPRM   :MAPS Terminating due to convergence"<<endl;
#endif
  
  if(fpref!=NULL) free(gfx_fp);
  status=MAPS_DONE;
  return true;//for now
}/*end RunMAPS*/
  
void maps::SetDefault(long V1, long P1, Bounds &B1, char* fn) {
/****
INPUT: The evaluation budget, dimension of space, bounds and function name.
EFFECT: This will automagically take care of generating an InitialDesign,
PSGrid, Search Criterion and a PatternSearch.  Life couldn't be easier. Sets
status=READY.  This gives you SearchCriterion's default, Latin hypercube
initial design and a CompassSearch.
****/ 
  V=V1;P=P1;b=B1;  
  Fname=NULL;
  if (ResetFunction(fn)) status=READY;
  else {
#if DEBUG>=1
    cerr<<"MPSD   :Executable File is not valid\n";
#endif
    status=NOTINIT;
  }/*end else*/

  C=0;x_c=0;
  N=V1/3; /*This is COMPLETELY arbitrary*/
  Eval_c=NULL;
  f_values=NULL;
  g=new PSGrid(P,b);
  ID=new LatinHCDesign(P);
  SC=new SearchCriterion(P,B1);    
  mp_vector temp(1); temp=0;
  PS=new CompassSearch(P,temp);
  (*ID).GenerateDesign(b,N);
}/*end SetDefault*/

bool maps::PCSubset(const pt_collect &A, const pt_collect &B) {
/****
INPUT: Two collections of points, with each point forming a row vector
OUTPUT: Returns true if A is a subset of B.
****/
  long AD=A.num_rows();
  long BD=B.num_rows();
  bool s=true;
  
  if(BD<AD) return false;
 
  for(long i=0;s&&i<AD;i++) {
    s=false;
    s=IsElement(A.row(i),B);
  }/*end for*/
  return s;
}/*end PCSubset*/


bool maps::IsElement(const mp_vector &A, const pt_collect &B) {
/****
INPUT: A point, and a collection of points.
OUTPUT: Returns true if A is in B.
****/
  long BD=B.num_rows();
  bool s=false;
  for(long j=0;!s&&j<BD;j++) 
    if(isnear(A,B.row(j),TOLERANCE/2.0)) s=true;  
  return s;
}/*end IsElement*/
 

double maps::EvalF(const mp_vector &point, bool &success) {
/****
INPUT: The point at which to evaluate the function, a boolean flag.
OUTPUT: The function value.
EFFECT: This function will do the fork/exec dance, sending the program the
parameters on its stdin, then after waiting, it will get a double from the
program's stdout.  If the program exited with an error, it will return ERROR and
set success to false.  Otherwise success will be true.
****/

#ifdef FUNCTION_COMPILED_MODE
  double f;
  success=(bool)mfcn(P,point.begin(),&f);
  return f;
#else    
  int childpid, pipe1[2], pipe2[2];
  long i;
  double ret_val;
  FILE *readpipe;
  FILE *writepipe;
  
  int state;
  
  if ((pipe(pipe1) < 0) || (pipe(pipe2) < 0) ) {
    perror("pipe");
    exit(-1);
  }/*end if*/

  if ((childpid = fork()) < 0) {
    perror("fork");
    exit(-1);
  } else if (childpid > 0) {   /*Parent*/
    close(pipe1[0]); close(pipe2[1]);
    /* Write to child on pipe1[1], read from child on pipe2[0]. */
    readpipe=fdopen(pipe2[0],"r");
    writepipe=fdopen(pipe1[1],"w");
    if(writepipe==NULL || readpipe==NULL)
      {perror("fdopen");cerr<<"MPEF   : Cannot open pipes... exiting\n";abort();}

    /*Output the vector x*/
    for(i=0;i<P;i++) fprintf(writepipe,"%1.8f ", point[i]); 
    fprintf(writepipe,"\n");
    fflush(writepipe);//repace setlinebuf
    
    /*Read in the returned value, and WAIT*/
    char c='a';
    int power_=0;
    int rv=0;
    fscanf(readpipe,"%lf", &ret_val);
    rv=fscanf(readpipe,"%c%d",&c,&power_);
    if(rv==EOF || (rv ==1) ) {}
    else{
      ret_val=ret_val * pow(10.0,power_);
    }
      
    wait(&state);

    /*Close outstanding pipes*/
    fclose(writepipe);fclose(readpipe);
    
    if(WIFEXITED(state)&&(WEXITSTATUS(state)==0)) {
      /*normal exit - no abort, return code 0*/
      success=true;
      return ret_val;     
    }/*end if*/
    else {
      success=false;
      return ERROR;
    }/*end else*/
    
  } else { /*Child*/
    close(pipe1[1]); close(pipe2[0]);
    /* Read from parent on pipe1[0], write to parent on pipe2[1]. */
    dup2(pipe1[0],0); dup2(pipe2[1],1);
    close(pipe1[0]); close(pipe2[1]);

    if (execlp(Fname, Fname, NULL) < 0) {
      perror("execlp");
      abort();
    }
    return ERROR;
    /* Never returns */
  }/*end else*/

  return ERROR;  /*never returns*/
#endif
}/*end EvalF*/





