/****
maps class - 8/2/00 cmsief
This IS the Model-Assisted Pattern Search.

To run with default settings (Latin Hypercube, Kriging Approximation, and
Dynamic Weighting Scheme):
call the constructor maps(long V1, long P1, const Bounds &B1, char*fn) ,
call RunMaps,
call RetrieveBestSeen.

NOTE: If not using defaults
Make sure the Grid and Pattern Search Core Patterns are compatible!
I am not responsible for what this might do if you don't.

FLAGS:
#define ITERATE_DEBUG - to output statistics iteration by iteration.
#define FUNCTION_COMPILED_MODE - to use the testClassic interface.

****/
#ifndef _MAPS_
#define _MAPS_
#include "maps_general.h"
#include "PSGrid.h"
#include "PatternSearch.h"
#include "InitialDesign.h"
#include "LatinHCDesign.h"
#include "SearchCriterion.h"
#include "CompassSearch.h"

/*Status flags for MAPS*/
#define NOTINIT 0    /* Not initialized yet*/
#define READY 1      /* Ready to run. Can still reset parameters */
#define MAPS_RUNNING 2  /* Running. Can't reset parameters */
#define MAPS_DONE 3  /* Finished. Can rerun, generate histories */

class maps {

public:

  maps();
  /*status=NOTINIT*/
  
  maps(long V1, long N1, long P1, const Bounds &B1, char* fn,
       const PSGrid &G1, const InitialDesign &I, const SearchCriterion &S1,
       const PatternSearch &PS1);
  /*Prep work and call ResetMAPSState*/
  maps(long V1, long P1, Bounds &B1, char*fn);
  /*Prep work and call SetDefault*/
  maps(maps& old);
  ~maps();

  /*****RESET FUNCTIONS****/
  bool ResetMAPSState(long V1, long N1, long P1, const Bounds &B1,
                      char* fn, const PSGrid &G1, const InitialDesign &I1,
                      const SearchCriterion &S1, const PatternSearch &PS1);
  /****
  INPUT: A partial set of state parameters for MAPS
  OUTPUT: Returns true if the operation was successful.
  EFFECT: Sets all the parameters, and sets status=READY,
  clears out Eval_c and f_values.  This wipes out everything else entirely.
  *****/

  void NewMAPSRun();
  /****
  EFFECT: If status=MAPS_DONE, clears the Eval_c and F_values,
  resets C=0 and sets status=READY.
  This allows you to reset MAPS between runs on a given function.
  ****/ 

  bool NewInitialPoint(const mp_vector& newpoint);
  /****
  INPUT: An initial guess for the starting point
  OUTPUT: True if the initial point was set. False on error.
  EFFECT: If the point is of the right dimension and in bounds,
  it will be used and true will be returned. 
  ****/
  
  bool ResetGrid(const PSGrid &G1);
  /****
  INPUT: A PSGrid
  OUTPUT: If it copies G1 it returns true, else it returns false.
  EFFECT: If status==READY, it copies G1 to G, else no effect.  
  ****/
  
  bool ResetDesign(const InitialDesign &I);
  /****
  INPUT: An InitialDesign.
  OUTPUT: If it copies I it returns true, else it returns false.
  EFFECT: If status==READY and I.IsDesignInitialized()==true,
  it copies I to ID, else no effect.  
  ****/

  bool ResetSearchCriterion(const SearchCriterion &S);
  /****
  INPUT: A SearchCriterion.
  OUTPUT: If it copies S it returns true, else it returns false.
  EFFECT: If status==READY, it copies S to SC, else no effect.
  ****/

  bool ResetPatternSearch(const PatternSearch &PS1);
  /****
  INPUT: A PatternSearch object.
  OUTPUT: If it copies S it returns true, else it returns false.
  EFFECT: If status==READY, it copies PS1 to PS, else no effect.
  ****/

  bool ResetFunction(char* fn);
  /****
  INPUT: String containing the name of the function to be optimized.
  OUTPUT: Returns true.
  EFFECT: Fname gets fn. 
  ****/

  /*****QUERY FUNCTIONS*****/
  
  bool RetrieveBestSeen(mp_vector* bpoint, double* bvalue) const;
  /****
  INPUT: Two Pointers, one for a mp_vector, one for a value
  OUTPUT: True if status==DONE, false if otherwise.
  EFFECT: If status=DONE, it finds the best point and its value and
  assigns it to the pointers.  This is done using a linear search.
  ****/
  
  void debug(char* code) const;
  /****
  EFFECT: Does a complete state dump to stdout.  This should do state dumps of
  all components.
  *****/       
  
  void GenerateMAPSGraphics(char* fpref) const;
  /****
  INPUT: Filename prefix.
  EFFECT: Outputs to fpref.MAPS.dat stuff that would be nice in
  producing pictures via GNUplot.  Due to limits in visualization
  technology, this feature is only available for functions from R or
  R^2.

  MAPS Manual, Section B explains how to generate graphics with gnuplot
  ****/

  void GenerateIterationHistory() const;
  /****
  EFFECT: Outputs the chosen points, their values and the best point and values
  returned to stdout.  To convert these data files into LaTeX tables with ease,
  use the included mdat2tex.awk script with the following syntax :
  awk -f mdat2tex.awk <mapsdatafile>
  awk will output the LaTeX to stdout.
  ****/
  
  /*****ACTION FUNCTIONS*****/
  bool RunMAPS(){return RunMAPS(NULL);}
  
  bool RunMAPS(char * fpref);
  /****
  INPUT: a pointer that specifies the filename prefix for
  iteration-by-iteration graphics.  Set this to NULL for no graphics,
  or call the parameter-free version above.
  OUTPUT: Returns true if nothing crashed.
  EFFECT: Assume status=READY.  Then it evaluates the function at the initial
  design sites. Then it begins the iterative process of looping until we use up
  the evaluation budget. When it finishes, it sets status=MAPS_DONE.  This will
  call EvalF.  If fpref!=NULL, then this will output graphics to files named :
  <fpref><C>.MAPS.AP.dat and <fpref><C>.MAPS.SC.dat .
  ****/       

private:

  void SetDefault(long V1, long P1, Bounds &B1, char*fn);
  /****
  INPUT: The evaluation budget, dimension of space, bounds and function name.
  EFFECT: This will automagically take care of generating an InitialDesign,
  PSGrid, Search Criterion and a PatternSearch.  Life couldn't be easier. Sets
  status=READY.  This gives you SearchCriterion's default, Latin Hypercube
  initial design and a CompassSearch.
  ****/ 

  static bool PCSubset(const pt_collect &A, const pt_collect &B);
  /****
  INPUT: Two collections of points, with each point forming a row vector
  OUTPUT: Returns true if A is a subset of B.
  ****/
 
  static bool IsElement(const mp_vector &A, const pt_collect &B);
  /****
  INPUT: A point, and a collection of points.
  OUTPUT: Returns true if A is in B.
  ****/
  
  double EvalF(const mp_vector& point, bool &success);
  /****
  INPUT: The point at which to evaluate the function, a boolean flag.
  OUTPUT: The function value.
  EFFECT: This function will do the fork/exec dance, sending the program the
  parameters on its stdin, then after waiting, it will get a double from the
  program's stdout.  If the program exited with an error, it will return
  ERROR and set success to false.  Otherwise success will be true.
  ****/
  
  long V;                /*Evaluation Budget*/
  long N;                /*Initial Design Allotment*/
  long C;                /*Current Point Number*/
  long P;                /*Dimension of the Space*/
  Bounds b;              /*Problem Bounds*/
  PSGrid *g;             /*The Grid - a real matrix*/
  pt_collect* Eval_c;    /*List of Evaluated Points - stored by row */  
  mp_vector* f_values;   /*List of Function Values*/
  char* Fname;           /*File Name of function to be optimized*/
  int status;            /*See above #defines*/
  long x_c;         /*The x_c'th element of Eval_c contains the current point*/
  InitialDesign* ID;     /*Initial Design*/
  SearchCriterion* SC;   /*Search Criterion*/
  PatternSearch* PS;     /*This will take care of optimizing S_c*/

  mp_vector* initial_guess; /*Initial starting point for MAPS*/

};/*end class*/

/*The objective function which you define elsewhere*/
extern int mfcn(long DIM, double * x, double *f);
#endif

